using UnityEngine;
using System.Collections;

public class EnemyController : MonoBehaviour {

	public float speed = 10;
	public string characterName;
	public Stats stats;
	public tk2dAnimatedSprite sprite;
	public float weight = 1f;
	public bool stationary = false;
	public bool gravity = true;
	public bool immortal = false;
	public GameObject messageDelegate = null;
	public string hitMessage = "";
	public string deathMessage = "";
	public bool behaviorEnabled = true;
	public float attackDamage = 5.0f;
	public Vector2 attackKnockback = new Vector2(20f,20f);
	public bool alwaysActive = false;
	protected float immunity = 0.0f;
	protected Transform _transform;
	protected Vector3 _position;
	protected Vector2 velocity = Vector3.zero;
	protected bool walking = false;
	protected float direction = -1f;
	protected bool flipped = false;
	
	public float x {
		get {
			return _position.x;
		}
		set {
			_position.Set(value, _position.y, _position.z);
		}
	}
	
	public float y {
		get {
			return _position.y;
		}
		set {
			_position.Set(_position.x, value, _position.z);
		}
	}
	
	public bool IsDead {
		get {
			return (stats.health <= 0);
		}
	}
	
	public bool IsImmune {
		get {
			if (immunity > 0) {
				return true;
			}
			return false;
		}
	}
	
	void Awake()
	{
		_transform = this.transform;
		_position = _transform.position;
	}
	
	// Use this for initialization
	void Start()
	{
	
	}
	
	// Update is called once per frame
	void Update()
	{
		
		CheckImmunity();
		CheckPhysics();
		
		if (behaviorEnabled) {
			if (alwaysActive || renderer.isVisible) {
				CheckDirection();
			}
		}
				
		SnapToGrid();
		
	}
	
	protected void CheckImmunity()
	{
		
		if (immunity > 0) {
			immunity -= Time.deltaTime;
			sprite.color = Color.Lerp(Color.white, Color.red, Mathf.PingPong(immunity * 19, 1f));
		} else {
			
			sprite.color = Color.white;
			
			if (this.IsDead) {
				Die();
			}
			
		}
		
	}
	
	protected virtual void CheckPhysics()
	{
		
		bool grounded = IsGrounded();
		
		//If We're Not Grounded 
		if (!grounded || velocity.y > 0f) {
			
			this.y += velocity.y * Time.deltaTime;
			
			if (gravity) {
				velocity.y += Game.gravity.y * Time.deltaTime;
			}
			
		} else {
			velocity.y = 0f;
		}
		
		if (!IsTouchingWall()) {
			
			if (velocity.x != 0) {
				this.x += velocity.x * Time.deltaTime;
			} else if (walking && immunity <= 0) {
				this.x += speed * direction * Time.deltaTime;
			}
			
		}
		
	}
	
	public float DistanceToPlayer()
	{
		
		PlayerController player = PlayerController.instance;
		
		Vector3 pt1 = _transform.position;
		Vector3 pt2 = player.transform.position;
		pt1.z = 0;
		pt2.z = 0;
		
		return Vector3.Distance(pt1, pt2);
		
	}
	
	public virtual void Die()
	{
		
		if (messageDelegate != null && deathMessage != "") {
			messageDelegate.SendMessage(deathMessage);
		}
		
		Destroy(gameObject);
		
	}
	
	public void ApplyDamage(float attack, Vector2 knockback)
	{
		
		if (immunity > 0) {
			return;
		}
		
		immunity = 0.3f;
		
		if (!stationary) {
			velocity += knockback;
		}
		
		if (!immortal) {
			stats.health -= attack;
		}
		
		if (messageDelegate != null && hitMessage != "") {
			messageDelegate.SendMessage(hitMessage);
		}
		
	}
	
	protected bool IsGrounded()
	{
		
		LayerMask mask = 1 << LayerMask.NameToLayer("Default");
		mask |= 1 << LayerMask.NameToLayer("Entity");
		
		RaycastHit hitInfo;
		Vector3 start = _transform.position;
		start.y += 6f;
		
		//Raycast to check for ground hit
		if (Physics.Raycast(start, Vector3.down, out hitInfo, 7.0f, mask.value)) {
			
			//Snap to ground if falling
			if (velocity.y <= 0) {
				velocity.y = 0;
				this.y = hitInfo.point.y;
			}
			
			if (immunity <= 0 && velocity.x != 0) {
				velocity.x = 0;
			}
			
			return true;
			
		}
		
		return false;
		
	}
	
	protected bool IsTouchingWall()
	{
		
		LayerMask mask = 1 << LayerMask.NameToLayer("Default");
		//mask |= 1 << LayerMask.NameToLayer("Entity");
		
		RaycastHit hitInfo;
		Vector3 start = _transform.position;
		start.y += 6f;
		
		//Raycast to check for ground hit
		if (Physics.Raycast(start, new Vector3(direction, 0, 0), out hitInfo, 8.0f, mask.value)) {
			return true;		
		}
		
		//If Direction and Velocity have opposite signs check the opposite direction too
		if (velocity.x != 0) {
			//If Direction and Velocity have opposite signs check the opposite direction too
			if ((direction > 0) != (velocity.x > 0)) {
				if (Physics.Raycast(_transform.position, new Vector3(direction * -1, 0, 0), out hitInfo, 8.0f, mask.value)) {
					return true;		
				}
			}
		}
		
		return false;
		
	}
	
	protected void SnapToGrid()
	{
		/*Vector3 pt = _position;
		pt.x = Mathf.Round(pt.x);
		pt.y = Mathf.Round(pt.y);
		_transform.position = pt;*/
		_transform.position = _position;
	}
	
	public virtual void CheckDirection()
	{
		
		if (PlayerController.instance == null) {
			return;
		}
		
		float playerX = PlayerController.instance.x;
		float dist = Mathf.Abs(playerX - this.x);
		
		if (stationary) {
			walking = false;
		} else if (dist > 2f) {
			walking = true;
		} else {
			walking = false;
		}
		
		if (playerX < this.x) {
			direction = -1f;
		} else if (playerX > this.x) {
			direction = 1f;
		}
		
		if (flipped && direction == -1.0f) {
			sprite.FlipX();
			flipped = !flipped;
		} else if (!flipped && direction == 1.0f) {
			sprite.FlipX();
			flipped = !flipped;
		}
		
	}
	
	public virtual void OnCollisionEnter(Collision collision)
	{
		//Check if hit player
		PlayerController player = collision.gameObject.GetComponent<PlayerController>();
		
		//Only do attack if no immunity
		if (player != null && immunity <= 0) {
			
			immunity = 0.3f;
			
			if (!stationary) {
				velocity.x += direction * -25;
				velocity.y += 25f;
			}
			
			player.ApplyDamage(attackDamage, attackKnockback);
		
		}
		
		Damage damage = collision.gameObject.GetComponent<Damage>();
		
		if (damage != null && immunity <= 0) {			
			this.ApplyDamage(damage.attack, damage.knockback);
			immunity = 0.3f;
		
		}
		
	}
	
	public IEnumerator WaitDialog(string text)
	{
		yield return StartCoroutine(GameUI.instance.WaitDialog("[#000000]" + characterName + ":[#FFFFFF] " + text));
	}
	
	public void AttachCamera()
	{
		//Attach Main Camera
		Camera.mainCamera.gameObject.GetComponent<CameraFollow>().target = this.transform;
	}
	
}
