/*
 *	LevelLoader.cs
 *
 *	Author: Cory R. Leach
 *  E-Mail: cory@mission-one.jp
 *
 *	Handles transitions between levels
 *
 */

using UnityEngine;
using System.Collections;

[AddComponentMenu("Kasetsu/LevelLoader")]

public class LevelLoader : MonoBehaviour {
	
	public float fadeDuration = 0.5f;
	public bool loadCleanLevel = true;
	
	public bool limitOneInstance = true;
	public Texture2D fadeTexture;
	
	public Texture2D[] loadingAnimation;
	
	//public bool testGui = false;
	
	private bool loadingNextLevel;
	private bool loadingCleanLevel;
	
	private string levelName = null;
	private int levelIndex = -1;
	
	private enum FadeStatus {Open, Closed, FadeIn, FadeOut, };
	private FadeStatus status = FadeStatus.Open;
	private GameObject fadeCurtain = null;
	private string previousLevel = null;
	
	private GameObject loadingAnimationObj = null;
	private int animationIndex = 0;
	
	private Color clearWhite = new Color(0.5f,0.5f,0.5f,0);
	//private Color clearBlack = new Color(0,0,0,0);
	
	static private LevelLoader sharedInstance = null;
	
	/*
	 * 
	 * Static Shared Instance Convinience Methods
	 * 
	 */
	static public LevelLoader SharedInstance() {
		return sharedInstance;
	}
	
	static public void LoadLevel(string name) {
	
		if ( sharedInstance == null) {
			Debug.Log("Not loading level because shared instance is null.");
			return;
		}
		
		sharedInstance._LoadLevel(name);
	
	}
	
	static public void LoadLevelAfterDelay(string name, float delay) {
	
		if ( sharedInstance == null) {
			Debug.Log("Not loading level because shared instance is null.");
			return;
		}
		
		sharedInstance._LoadLevelAfterDelay(name,delay);
		
	}
	
	static public void FadeInScene() {
	
		if ( sharedInstance == null) {
			return;
		}
		
		sharedInstance._FadeInScene();
	
	}
	
	static public void LoadPreviousLevel() {
	
		if ( sharedInstance == null && sharedInstance.previousLevel != null ) {
			return;
		}
		
		sharedInstance._LoadLevel(sharedInstance.previousLevel);
		
	}
	
	static public void FadeOutScene()
	{
	
		if (sharedInstance == null) {
			return;
		}
		
		sharedInstance._FadeOutScene();
	
	}
	
	static public void FadeOutScene(Color color)
	{
	
		if (sharedInstance == null) {
			return;
		}
		
		sharedInstance._FadeOutScene(color);
	
	}
	
	/*
	 * 
	 * Instance Methods
	 * 
	 */
	void Awake()
	{
		
		if (sharedInstance != null) {
			Destroy(gameObject);
		}
		
		loadingNextLevel = false;
		
	}
	
	void Start()
	{
			
		//If shared instance doesn't exist, set this as shared instance
		if (sharedInstance == null) {
			
			sharedInstance = this;
			DontDestroyOnLoad(gameObject);
			
			if (fadeTexture != null) {
				fadeCurtain = iTween.CameraFadeAdd(fadeTexture);
			} else {
				fadeCurtain = iTween.CameraFadeAdd();
			}
			
			fadeCurtain.transform.position = new Vector3(0.5f, 0.5f, 0f);
			fadeCurtain.layer = LayerMask.NameToLayer("FadeCurtain");
			
			if (loadingAnimation.Length > 0) {
				loadingAnimationObj = new GameObject("LoadingAnimation");
				loadingAnimationObj.AddComponent<GUITexture>();
				loadingAnimationObj.layer = LayerMask.NameToLayer("GUI");
				loadingAnimationObj.transform.position = new Vector3(0.5f, 0.1f, 1f);
				loadingAnimationObj.transform.localScale = new Vector3(0f, 0f, 0f);
				loadingAnimationObj.guiTexture.pixelInset = new Rect(692f * -0.5f, 273f * -0.5f, 692f, 273f);
				loadingAnimationObj.guiTexture.color = new Color(1.0f, 1.0f, 1.0f, 0f);
				loadingAnimationObj.guiTexture.texture = loadingAnimation[0];
				DontDestroyOnLoad(loadingAnimationObj);
			}
			
			
			DontDestroyOnLoad(fadeCurtain);
		
		} else if (sharedInstance != this && limitOneInstance) {
			Destroy(gameObject);
		}
		
	}
	
	int frame = 0;
	int direction = 1;	
	float frameDuration = 0.1f;
	float frameTime = 0f;
	void Update()
	{

		if (loadingAnimationObj == null) {
			return;
		}
		
		if (loadingAnimationObj.guiTexture.color.a > 0) {

			frameTime += Time.deltaTime;

			if (frameTime > frameDuration) {

				if (frame == 0) {
					direction = 1;
				} else if (frame == loadingAnimation.Length - 1) {
					direction = -1;
				}
				
				frame = (frame + direction);
				
				loadingAnimationObj.guiTexture.texture = loadingAnimation[frame];
				frameTime = 0;

			}

		}

	}
	
	void OnLevelWasLoaded(int index)
	{
		
		if (sharedInstance != this) {
			return;
		}
		
		Debug.Log("Loaded Level " + index + " : " + Application.loadedLevelName);
		
		if (loadingCleanLevel) {
			
			Debug.Log("Clean level Loaded");
			//Force unloading of assets
			Resources.UnloadUnusedAssets();
			System.GC.Collect();
			
			//Load Level
			if (levelName != null) {
				loadingNextLevel = true;
				Debug.Log("LWL Calling Load Level " + levelName);
				//Application.LoadLevel(levelName);
				Application.LoadLevelAsync(levelName);
			} else if (levelIndex != -1) {
				loadingNextLevel = true;
				Debug.Log("LWL Calling Load Level " + levelIndex);
				//Application.LoadLevel(levelIndex);
				Application.LoadLevelAsync(levelIndex);
			} else {
				//This is an error state
				//No Level To Load Specified, fade back in immediately
				Debug.LogError("No Level to Load Specified!");
				_FadeInScene();
			}
			
			levelName = null;
			levelIndex = -1;
			loadingCleanLevel = false;
			
		} else if (loadingNextLevel) {
		
			//Next Level Loaded, now fade in scene
			_FadeInScene();
			Debug.Log("Level Was Loaded, now fade in scene");
			loadingNextLevel = false;
			//Invoke("_FadeInScene", 0.1f);
			
		}
		
	}
	
	void _LoadLevelAfterDelay(string name, float delay) {
		StartCoroutine(_LoadLevelAfterDelayCoroutine(name,delay));
	}
	
	IEnumerator _LoadLevelAfterDelayCoroutine(string name, float delay) {
	
		yield return new WaitForSeconds(delay);
		
		_LoadLevel(name);
	
	}
	
	void _LoadLevel(string name) {
		
		previousLevel = Application.loadedLevelName;
		
		if ( loadingCleanLevel || loadingNextLevel ) {
			return;
		}
		
		Debug.Log("Loading Level " + name);
		
		if ( loadCleanLevel ) {
			loadingCleanLevel = true;
			loadingNextLevel = false;
		} else {
			loadingCleanLevel = false;
			loadingNextLevel = true;
		}
		
		levelName = name;
		_FadeOutScene();
		
	}
	
	void _LoadLevel(int index) {
		
		previousLevel = Application.loadedLevelName;
		
		if ( loadingCleanLevel || loadingNextLevel ) {
			return;
		}
		
		if ( loadCleanLevel ) {
			loadingCleanLevel = true;
			loadingNextLevel = false;
		} else {
			loadingCleanLevel = false;
			loadingNextLevel = true;
		}
		
		levelIndex = index;
		_FadeOutScene();
		
	}
	
	void _FadeOutScene()
	{
		
		fadeCurtain.guiTexture.color = clearWhite;
		status = FadeStatus.FadeOut;
		iTween.CameraFadeTo(iTween.Hash("amount", 1.0f, "time", fadeDuration, "oncomplete", "FadeFinish", "oncompletetarget", gameObject));
		
		if (loadingAnimationObj != null) {
			iTween.FadeTo(loadingAnimationObj, iTween.Hash("amount", 1.0f, "time", fadeDuration));
		}
		//StartCoroutine(DoLoadingAnimation());
		
	}
	
	void _FadeOutScene(Color color)
	{
		color.a = 0.0f;
		fadeCurtain.guiTexture.color = color;
		
		if (loadingAnimationObj != null) {
			loadingAnimationObj.guiTexture.color = color;
		}
		
		status = FadeStatus.FadeOut;
		iTween.CameraFadeTo(iTween.Hash("amount", 1.0f, "time", fadeDuration, "oncomplete", "FadeFinish", "oncompletetarget", gameObject));
		
		if (loadingAnimationObj != null) {
			iTween.FadeTo(loadingAnimationObj, iTween.Hash("amount", 1.0f, "time", fadeDuration));
		}
		//StartCoroutine(DoLoadingAnimation());
		
	}
	
	void _FadeInScene()
	{
			
		status = FadeStatus.FadeIn;
		iTween.CameraFadeTo(iTween.Hash("amount", 0.0f, "time", fadeDuration, "oncomplete", "FadeFinish", "oncompletetarget", gameObject));
		
		if (loadingAnimationObj != null) {
			iTween.FadeTo(loadingAnimationObj, iTween.Hash("amount", 0.0f, "time", fadeDuration));
		}
		
	}
	
	/*IEnumerator DoLoadingAnimation()
	{
		
		int frame = 0;
		int direction = 1;
		float frameDuration = 0.1f;
		
		loadingAnimationObj.guiTexture.texture = loadingAnimation[frame];
		
		yield return StartCoroutine(Wait(0.03f));//new WaitForSeconds(0.03f);
		
		while (loadingAnimationObj.guiTexture.color.a > 0) {
		
			yield return StartCoroutine(Wait(frameDuration));
			
			if (frame == 0) {
				direction = 1;
			} else if (frame == loadingAnimation.Length - 1) {
				direction = -1;
			}
				
			frame = (frame + direction);
			
			loadingAnimationObj.guiTexture.texture = loadingAnimation[frame];
						
		}
		
	}

	private IEnumerator Wait(float p)
	{
		float pauseEndTime = Time.realtimeSinceStartup + 1;
		while (Time.realtimeSinceStartup < pauseEndTime) {
			yield return 0;
		}
	}*/
	
	private void FadeFinish()
	{
		
		if (status == FadeStatus.FadeOut) {
		
			//Set status to closed
			status = FadeStatus.Closed;
			
			//Check if we should load clean level
			if (loadingCleanLevel) {
				
				Application.LoadLevelAsync("Clean");
				//Return here so we don't clear out the level name and index
				return;
				
				//Check if we should load a level
			} else if (loadingNextLevel) {
				
				//Load Level
				if (levelName != null) {
					//Application.LoadLevel(levelName);
					Application.LoadLevelAsync(levelName);
				} else if (levelIndex != -1) {
					//Application.LoadLevel(levelIndex);
					Application.LoadLevelAsync(levelIndex);
				} else {
					//This is an error state
					//No Level To Load Specified, fade back in immediately
					Debug.LogError("No Level to Load Specified!");
					_FadeInScene();
				}
				
			}

		} else if (status == FadeStatus.FadeIn) {

			//Set status to open
			status = FadeStatus.Open;
		
		}
		
		//Reset Level Load State
		levelName = null;
		levelIndex = -1;
		
	}
	
}
