using UnityEngine;
using System.Collections;

public enum PlayerColor
{
	Blue,
	Red,
	Yellow,
	Purple,
	Green,
	Orange
}

[RequireComponent(typeof(Rigidbody))]
public class PlayerController : MonoBehaviour {
	
	public Stats stats {
		get {
			return Game.stats;
		}
	}
	public float moveSpeed = 32f;
	public float jumpVelocity = 64f;
	
	public float paintRegen = 0.5f;
	
	public Vector2 velocity = Vector2.zero;
	public float direction = 1f;
	
	public string characterName = "Blue";
	public string color = "[#0000FF]";
	
	//public tk2dSprite sprite;
	public tk2dAnimatedSprite sprite;
	public bool controlEnabled = true;
	private bool flipped = false;
	public float attackCooldown = 0f;
	
	public bool destroyOnLoad = false;
	
	public PlayerColor weaponColor = PlayerColor.Blue;
	
	public SpriteText actionIndicator;
	
	private float immunity = 0;
	
	public float y {
		get {
			return transform.position.y;
		}
		set {
			Vector3 pt = transform.position;
			pt.y = value;
			transform.position = pt;
		}
	}
	
	public float x {
		get {
			return transform.position.x;
		}
		set {
			Vector3 pt = transform.position;
			pt.x = value;
			transform.position = pt;
		}
	}
	
	public bool isDead {
		get {
			return (stats.health <= 0);
		}
	}
	
	private static PlayerController _instance;
	public static PlayerController instance {
		get {
			return _instance;
		}
	}
	
	void Awake()
	{
		
		if (_instance != null) {
			Destroy(gameObject);
			return;
		}
		
		_instance = this;
		if (!destroyOnLoad) {
			DontDestroyOnLoad(this.gameObject);
		}
		
		AttachCamera();
		
		if (ActionColorBlock.destinationOnLoad != "") {
			SetPositionAtDestination(ActionColorBlock.destinationOnLoad);
			ActionColorBlock.destinationOnLoad = "";
		}
		
	}
	
	IEnumerator OnLevelWasLoaded(int level)
	{
		
		if (ActionColorBlock.destinationOnLoad != "") {
			SetPositionAtDestination(ActionColorBlock.destinationOnLoad);
			ActionColorBlock.destinationOnLoad = "";
		}
		
		while (GameUI.instance == null) {
			yield return null;
		}
		
		AttachCamera();
		
	}
	
	void SetPositionAtDestination(string destinationObjectName)
	{
		GameObject gobj = GameObject.Find(destinationObjectName);
		
		if (gobj == null) {
			Debug.LogWarning("Destination Object Doesn't Exist");
			return;
		}
		
		Vector3 newPos = gobj.transform.position;
		newPos.z = this.transform.position.z;
		this.transform.position = newPos;
		
	}
	
	public void AttachCamera()
	{
		CameraFollow cameraFollow = Camera.mainCamera.gameObject.GetComponent<CameraFollow>();
		
		if (cameraFollow == null) {
			return;
		}
		
		//Attach Main Camera
		cameraFollow.target = this.transform;
	}
	
	// Use this for initialization
	void Start () 
	{
	
	}
	
	// Update is called once per frame
	void Update()
	{
		CheckImmunity();
		CheckInput();
		CheckPhysics();
		CheckAnimation();
		
		//Regen health if not dead
		if (!this.isDead) {
			if (stats.health < stats.maxHealth) {
				stats.health = Mathf.Clamp(stats.health + paintRegen * Time.deltaTime, 0, stats.maxHealth);
			}
		}
		
	}
	
	void Attack()
	{
		
		if (stats.health <= 1f) {
			return;
		}
		
		if (ParticleManager.Instance == null) {
			return;
		}
		
		stats.health -= 1f;
		attackCooldown = 1f / stats.attackSpeed;
		
		Particle particle = ParticleManager.Instance.SpawnParticle(transform.position + new Vector3(3f, 3f, 0), "paintparticle");
		particle.velocity = new Vector3(velocity.x + moveSpeed * direction + 40f * (flipped ? -1 : 1), velocity.y + 40f);
		
		particle.color = Game.GetColor(weaponColor);
		
		if (Game.GetFlag("ReachedMonochrome")) {
			particle.dieOnGroundHit = true;
		} else {
			particle.dieOnGroundHit = false;
		}
		
		/*Color[] colors = { Color.blue,
							Color.yellow,
							Color.red,
							Color.green,
							new Color(130.0f / 255.0f, 0f, 1.0f, 1.0f),
							new Color(1.0f, 123.0f / 255.0f, 0f, 1.0f)
		};
		
		particle.color = colors[Random.Range(0,colors.Length)];
		*/
		
	}
	
	void CheckAnimation()
	{
		
		if (direction < 0 && !flipped) {
			flipped = true;
			sprite.FlipX();
		} else if (direction > 0 && flipped) {
			flipped = false;
			sprite.FlipX();
		}
		
	}
	
	void CheckInput()
	{
		
		direction = GetHorizontalInput();
		
		bool grounded = IsGrounded();
		
		ActionColorBlock action = CheckAction();
		if (action != null && action.painted) {
			actionIndicator.gameObject.active = true;
		} else {
			actionIndicator.gameObject.active = false;
		}
		
		if (direction != 0 && grounded) {
			if (!sprite.Playing) {
				sprite.Play();
			}
		} else if (sprite.Playing) {
			sprite.Stop();
			sprite.SetFrame(0);
		}
		
		if (GetJumpInput() && IsGrounded()) {
			velocity.y = jumpVelocity;
		}
		
		if (attackCooldown > 0) {
			attackCooldown -= Time.deltaTime;
		} else if (GetFire()) {
			if (action != null && action.painted) {
				action.Action();
			} else {
				Attack();
			}			
		}
		
	}
	
	void CheckPhysics()
	{
		
		bool grounded = IsGrounded();
		
		//If We're Not Grounded 
		if (!grounded || velocity.y > 0f) {
			
			//Reduce Y Velocity to 0 if we hit the ceiling
			if (velocity.y > 0f && IsTouchingCeiling()) {
				velocity.y = 0;
			}
				
			this.y += velocity.y * Time.deltaTime;
			velocity.y += Game.gravity.y * Time.deltaTime;
			
		} else {
			velocity.y = 0f;
		}
		
		if (!IsTouchingWall()) {
		
			if (velocity.x != 0f) {
				this.x += velocity.x * Time.deltaTime;
			}
			
			this.x += moveSpeed * direction * Time.deltaTime;
			
		}
		
		//Die if we fall out of bounds
		if (this.y < -50f) {
			stats.health = 0;
		}
		
	}
	
	protected void CheckImmunity()
	{
		
		if (immunity > 0) {
			
			immunity -= Time.deltaTime;
			sprite.color = Color.Lerp(Color.white, Color.red, Mathf.PingPong(immunity * 19, 1f));
			
		} else {
			
			sprite.color = Color.white;
			
			if (this.isDead) {
				Die();
			}
			
		}
		
	}
	
	float GetHorizontalInput()
	{
		if (!controlEnabled) {
			return 0;
		}
		return Input.GetAxis("Horizontal");
	}
	
	bool GetFire()
	{
		if (!controlEnabled) {
			return false;
		}
		return Input.GetButtonDown("Fire1");
	}
	
	bool GetJumpInput()
	{
		if (!controlEnabled) {
			return false;
		}
		return Input.GetButtonDown("Jump");
	}
	
	ActionColorBlock CheckAction()
	{
				
		Vector3 start = transform.position;
		start.y += 8f;
		start.z = -10;
		
		RaycastHit[] hits = Physics.SphereCastAll(start, 8f, new Vector3(0, 0, 1f), 20.0f);
		
		foreach (RaycastHit hitInfo in hits) {
		
			ActionColorBlock action = hitInfo.collider.gameObject.GetComponent<ActionColorBlock>();
		
			if (action != null) {
				return action;
			}
			
		}
		
		return null;
		
	}
	
	bool IsGrounded()
	{
		
		LayerMask mask = (1 << LayerMask.NameToLayer("Default"));
		
		RaycastHit hitInfo;
		Vector3 start = transform.position + Vector3.up * 4f;
		
		//Raycast to check for ground hit
		if (Physics.SphereCast(start, 4f, Vector3.down, out hitInfo, 4.0f, mask.value)) { 
			
			//Snap to ground if falling
			if (velocity.y <= 0) {
				this.y = hitInfo.point.y;
				velocity.y = 0;
			}
			
			if (immunity <= 0 && velocity.x != 0) {
				velocity.x = 0;
			}
			
			return true;
			
		}
		
		return false;
		
	}
	
	bool IsTouchingWall()
	{
		
		LayerMask mask = 1 << LayerMask.NameToLayer("Default");
		//mask |= (1 << LayerMask.NameToLayer("Entity"));
		
		RaycastHit hitInfo;
		Vector3 start = transform.position;
		start.y += 8f;
		
		//Raycast to check for ground hit
		if (Physics.SphereCast(start, 6f, new Vector3(direction, 0, 0), out hitInfo, 1.0f, mask.value)) {
			return true;
		}
		
		//Check again at base of sprite
		/*if (Physics.Raycast(transform.position, new Vector3(direction, 0, 0), out hitInfo, 2.0f, mask.value)) {
			return true;
		}*/
		
		if (velocity.x != 0) {
			//If Direction and Velocity have opposite signs check the opposite direction too
			if ((direction > 0) != (velocity.x > 0)) {
				if (Physics.SphereCast(start, 6f, new Vector3(direction * -1, 0, 0), out hitInfo, 3.0f, mask.value)) {
					return true;
				}
			}
		}
		
		return false;
		
	}
	
	bool IsTouchingCeiling()
	{
		LayerMask mask = 1 << LayerMask.NameToLayer("Default");
		
		RaycastHit hitInfo;
		Vector3 pt = transform.position;
		pt.y += 8;
				
		//Raycast to check for ground hit
		if (Physics.Raycast(pt, new Vector3(0, 1, 0), out hitInfo, 8.0f, mask.value)) {
			
			//Hit ceiling if this is a paint block
			if (hitInfo.collider.gameObject.GetComponent<PaintBlock>() != null) {
				return true;		
			}
			
		}
		
		return false;
	}
	
	public IEnumerator WaitDialog(string text)
	{
		yield return StartCoroutine(GameUI.instance.WaitDialog(color + characterName + ":[#FFFFFF] " + text));
	}
	
	public void ApplyDamage(float attack, Vector2 knockback)
	{
		
		if (immunity > 0) {
			return;
		}
		
		immunity = 0.3f;
		
		velocity += knockback;
		
		stats.health -= attack;
		
	}
	
	public void Die()
	{
		
		ActionColorBlock.destinationOnLoad = "EnterFromDead";
		
		if (Game.GetFlag("ReachedCastle")) {
			
			DeadScript.sceneName = "Castle"; 
			
			if (Game.GetFlag("ReachedMonochrome")) {
				DeadScript.destinationOnLoad = "EnterFromMonochrome";
			} else {
				DeadScript.destinationOnLoad = "EnterFromMountain";
			}
						
		} else {
		
			DeadScript.destinationOnLoad = "EnterFromDead";
			DeadScript.sceneName = "BluesHouse"; 
			
		}
		
		stats.health = stats.maxHealth;
		LevelLoader.LoadLevel("Dead");
		
	}
	
	void OnCollisionEnter(Collision collision)
	{
		
		if (collision.gameObject.layer != LayerMask.NameToLayer("Enemy")) {
			return;
		}
		
		Damage damage = collision.gameObject.GetComponent<Damage>();
		
		if (damage != null && immunity <= 0) {
			Debug.Log("Damaged");
			
			this.ApplyDamage(damage.attack, damage.knockback);
			immunity = 0.3f;
		
		}
		
	}
	
	void OnCollisionStay(Collision collision)
	{
		if (collision.gameObject.layer != LayerMask.NameToLayer("Enemy")) {
			return;
		}
		
		Damage damage = collision.gameObject.GetComponent<Damage>();
		
		if (damage != null && immunity <= 0) {
			Debug.Log("Damaged");
			
			this.ApplyDamage(damage.attack, damage.knockback);
			immunity = 0.3f;
		
		}
	}
	
	
}
